@@ -5,13 +5,14 @@ from .graph import Graph
 
 __all__ = ['Graph']
 
-__version__ = '0.7'
-__author__ = 'Yu-Hang "Maxin" Tang, Oguz Selvitopi'
+__version__ = '0.8.1'
+__author__ = '''Yu-Hang "Maxin" Tang, Oguz Selvitopi, Chi-Feng Wang,
+    Thomas H. Li'''
 __maintainer__ = 'Yu-Hang "Maxin" Tang'
 __email__ = 'Tang.Maxin@gmail.com'
 __license__ = 'see LICENSE file'
 __copyright__ = '''
-GraphDot Copyright (c) 2019, The Regents of the University of California,
+GraphDot Copyright (c) 2019-2021, The Regents of the University of California,
 through Lawrence Berkeley National Laboratory (subject to receipt of any
 required approvals from the U.S. Dept. of Energy).  All rights reserved.
 
