@@ -212,9 +212,9 @@ def get_ringlist(mol):
     return [sorted(rings) if len(rings) else [0] for rings in ringlist]
 
 
-def _from_rdkit(cls, mol, bond_type='order', set_ring_list=True,
+def _from_rdkit(cls, mol, title=None, bond_type='order', set_ring_list=True,
                 set_ring_stereo=True):
-    g = nx.Graph()
+    g = nx.Graph(title=title)
 
     for i, atom in enumerate(mol.GetAtoms()):
         g.add_node(i)
@@ -235,7 +235,7 @@ def _from_rdkit(cls, mol, bond_type='order', set_ring_list=True,
         if bond_type == 'order':
             g.edges[ij]['order'] = bond.GetBondTypeAsDouble()
         else:
-            g.edges[ij]['type'] = bond.GetBondTypeAsDouble()
+            g.edges[ij]['type'] = bond.GetBondType()
         g.edges[ij]['aromatic'] = bond.GetIsAromatic()
         g.edges[ij]['conjugated'] = bond.GetIsConjugated()
         g.edges[ij]['stereo'] = bond.GetStereo()
