@@ -2,12 +2,14 @@
 # -*- coding: utf-8 -*-
 import copy
 import numbers
+import warnings
 import itertools as it
 from collections import namedtuple
 import numpy as np
 from graphdot.graph import Graph
 from graphdot.util import Timer
-from graphdot.util.iterable import fold_like, flatten
+from graphdot.util.iterable import fold_like, flatten, replace
+from graphdot.util.pretty_tuple import pretty_tuple
 from ._backend_factory import backend_factory
 from .starting_probability import StartingProbability, Uniform, Adhoc
 
@@ -24,13 +26,24 @@ class MarginalizedGraphKernel:
         A kernelet that computes the similarity between individual nodes
     edge_kernel: microkernel
         A kernelet that computes the similarity between individual edge
-    kwargs: optional arguments
-        p: positive number (default=1.0) or :py:class:`StartingProbability`
-            The starting probability of the random walk on each node. Must be
-            either a positive number or a concrete subclass instance of
-            :py:class:`StartingProbability`.
-        q: float in (0, 1)
-            The probability for the random walk to stop during each step.
+    p: positive number (default=1.0) or :py:class:`StartingProbability`
+        The starting probability of the random walk on each node. Must be
+        either a positive number or a concrete subclass instance of
+        :py:class:`StartingProbability`.
+    q: float in (0, 1)
+        The probability for the random walk to stop during each step.
+    q_bounds: pair of floats
+        The lower and upper bound that the stopping probability can vary during
+        hyperparameter optimization.
+    eps: float
+        The step size used for finite difference approximation of the gradient.
+        Only used for nodal matrices (``nodal=True``).
+    dtype: numpy dtype
+        The data type of the kernel matrix to be returned.
+    backend: 'auto' or 'cuda' or an instance of
+    :py:class:`graphdot.kernel.marginalized.Backend`.
+        The computing engine that solves the marginalized graph kernel's
+        generalized Laplacian equation.
     """
     trait_t = namedtuple(
         'Traits', 'diagonal, symmetric, nodal, lmin, eval_gradient'
@@ -42,24 +55,42 @@ class MarginalizedGraphKernel:
         traits = cls.trait_t(
             diagonal, symmetric, nodal, lmin, eval_gradient
         )
-        if traits.eval_gradient is True:
-            if nodal is not False:
-                raise ValueError(
-                    'Gradients can only be evaluated with nodal=False'
-                )
         return traits
 
     def __init__(self, node_kernel, edge_kernel, p=1.0, q=0.01,
-                 q_bounds=(1e-4, 1 - 1e-4), dtype=np.float, backend='auto'):
+                 q_bounds=(1e-4, 1 - 1e-4), eps=1e-2, ftol=1e-8, gtol=1e-6,
+                 dtype=np.float, backend='auto'):
         self.node_kernel = node_kernel
         self.edge_kernel = edge_kernel
         self.p = self._get_starting_probability(p)
         self.q = q
         self.q_bounds = q_bounds
+        self.eps = eps
+        self.ftol = ftol
+        self.gtol = gtol
         self.element_dtype = dtype
 
         self.backend = backend_factory(backend)
 
+        if self.node_kernel.minmax[0] <= 0 or self.node_kernel.minmax[1] > 1:
+            warnings.warn(
+                'Node kernel value range should be within (0, 1], '
+                f'got {self.node_kernel.minmax} for {self.node_kernel}. '
+                'This will not be allowed in a future version. '
+                'Consider adding a small constant or using the `.normalized` '
+                'attribute of the kernel.',
+                DeprecationWarning
+            )
+        if self.edge_kernel.minmax[0] < 0 or self.edge_kernel.minmax[1] > 1:
+            warnings.warn(
+                'Edge kernel value range must be within [0, 1], '
+                f'got {self.edge_kernel.minmax} for {self.edge_kernel}. '
+                'This will not be allowed in a future version. '
+                'Consider adding a small constant or using the `.normalized` '
+                'attribute of the kernel.',
+                DeprecationWarning
+            )
+
     def _get_starting_probability(self, p):
         if isinstance(p, StartingProbability):
             return p
@@ -178,9 +209,14 @@ class MarginalizedGraphKernel:
                 starts[:len(X)] = np.arange(len(X))
                 starts[len(X):] = np.arange(len(Y) + 1)
                 output_shape = (len(X), len(Y))
+        gramian = backend.empty(int(np.prod(output_shape)), np.float32)
         if traits.eval_gradient is True:
-            output_shape = (*output_shape, 1 + self.n_dims)
-        output = backend.empty(int(np.prod(output_shape)), np.float32)
+            gradient = backend.empty(
+                self.n_dims * int(np.prod(output_shape)), np.float32
+            )
+        else:
+            gradient = None
+
         timer.toc('creating output buffer')
 
         ''' call GPU kernel '''
@@ -191,10 +227,16 @@ class MarginalizedGraphKernel:
             self.edge_kernel,
             self.p,
             self.q,
+            self.eps,
+            self.ftol,
+            self.gtol,
             jobs,
             starts,
-            output,
-            output_shape,
+            gramian,
+            gradient,
+            output_shape[0],
+            output_shape[1],
+            self.n_dims,
             traits,
             timer,
         )
@@ -202,7 +244,11 @@ class MarginalizedGraphKernel:
 
         ''' collect result '''
         timer.tic('collecting result')
-        output = output.reshape(*output_shape, order='F')
+        gramian = gramian.reshape(*output_shape, order='F')
+        if gradient is not None:
+            gradient = gradient.reshape(
+                (*output_shape, self.n_dims), order='F'
+            )[:, :, self.active_theta_mask]
         timer.toc('collecting result')
 
         if timing:
@@ -211,13 +257,14 @@ class MarginalizedGraphKernel:
 
         if traits.eval_gradient is True:
             return (
-                output[:, :, 0].astype(self.element_dtype),
-                output[:, :, 1:].astype(self.element_dtype)
+                gramian.astype(self.element_dtype),
+                gradient.astype(self.element_dtype)
             )
         else:
-            return output.astype(self.element_dtype)
+            return gramian.astype(self.element_dtype)
 
-    def diag(self, X, eval_gradient=False, nodal=False, lmin=0, timing=False):
+    def diag(self, X, eval_gradient=False, nodal=False, lmin=0,
+             active_theta_only=True, timing=False):
         """Compute the self-similarities for a list of graphs
 
         Parameters
@@ -240,6 +287,9 @@ class MarginalizedGraphKernel:
             of Eq. 1 in Tang & de Jong, 2019 https://doi.org/10.1063/1.5078640
             (or the first unnumbered equation in Section 3.3 of Kashima, Tsuda,
             and Inokuchi, 2003).
+        active_theta_only: bool
+            Whether or not to return only gradients with regard to the
+            non-fixed hyperparameters.
 
         Returns
         -------
@@ -299,12 +349,13 @@ class MarginalizedGraphKernel:
             np.cumsum(sizes**2, out=starts[1:])
             output_length = int(starts[-1])
         else:
-            raise(ValueError("Invalid 'nodal' option '%s'" % nodal))
+            raise ValueError("Invalid 'nodal' option '%s'" % nodal)
+        gramian = backend.empty(output_length, np.float32)
         if traits.eval_gradient is True:
-            output_shape = (output_length, 1 + self.n_dims)
+            gradient = backend.empty(self.n_dims * output_length, np.float32)
         else:
-            output_shape = (output_length, 1)
-        output = backend.empty(int(np.prod(output_shape)), np.float32)
+            gradient = None
+
         timer.toc('creating output buffer')
 
         ''' call GPU kernel '''
@@ -315,10 +366,16 @@ class MarginalizedGraphKernel:
             self.edge_kernel,
             self.p,
             self.q,
+            self.eps,
+            self.ftol,
+            self.gtol,
             jobs,
             starts,
-            output,
-            output_shape,
+            gramian,
+            gradient,
+            output_length,
+            1,
+            self.n_dims,
             traits,
             timer,
         )
@@ -326,17 +383,22 @@ class MarginalizedGraphKernel:
 
         ''' post processing '''
         timer.tic('collecting result')
-        output = output.reshape(*output_shape, order='F')
+        if gradient is not None:
+            gradient = gradient.reshape(
+                (output_length, self.n_dims), order='F'
+            )
+            if active_theta_only:
+                gradient = gradient[:, self.active_theta_mask]
         if nodal == 'block':
-            retval = [output[s:s + n**2].reshape(n, n)
+            retval = [gramian[s:s + n**2].reshape(n, n)
                       for s, n in zip(starts[:-1], sizes)]
         elif traits.eval_gradient is True:
             retval = (
-                output[:, 0].ravel().astype(self.element_dtype),
-                output[:, 1:].astype(self.element_dtype)
+                gramian.astype(self.element_dtype),
+                gradient.astype(self.element_dtype)
             )
         else:
-            retval = output.ravel().astype(self.element_dtype)
+            retval = gramian.astype(self.element_dtype)
         timer.toc('collecting result')
 
         if timing:
@@ -354,59 +416,93 @@ class MarginalizedGraphKernel:
     def requires_vector_input(self):
         return False
 
-    @property
-    def n_dims(self):
-        '''p.theta + q + node_kernel.theta + edge_kernel.theta'''
-        return len(self.theta)
-
     @property
     def hyperparameters(self):
-        return namedtuple(
-            'GraphKernelHyperparameters',
-            ['p', 'q', 'node', 'edge']
+        '''A hierarchical representation of all the kernel hyperparameters.'''
+        return pretty_tuple(
+            'MarginalizedGraphKernel',
+            ['starting_probability', 'stopping_probability', 'node_kernel',
+             'edge_kernel']
         )(self.p.theta,
           self.q,
           self.node_kernel.theta,
           self.edge_kernel.theta)
 
     @property
-    def theta(self):
-        return np.log(np.fromiter(flatten(self.hyperparameters), np.float))
-
-    @theta.setter
-    def theta(self, value):
-        (self.p.theta,
-         self.q,
-         self.node_kernel.theta,
-         self.edge_kernel.theta
-         ) = fold_like(np.exp(value), self.hyperparameters)
+    def flat_hyperparameters(self):
+        return np.fromiter(flatten(self.hyperparameters), np.float)
 
     @property
     def hyperparameter_bounds(self):
-        return namedtuple(
+        return pretty_tuple(
             'GraphKernelHyperparameterBounds',
-            ['p', 'q', 'node', 'edge']
+            ['starting_probability', 'stopping_probability', 'node_kernel',
+             'edge_kernel']
         )(self.p.bounds,
           self.q_bounds,
           self.node_kernel.bounds,
           self.edge_kernel.bounds)
 
+    @property
+    def n_dims(self):
+        'Number of hyperparameters including both optimizable and fixed ones.'
+        return len(self.flat_hyperparameters)
+
+    @property
+    def active_theta_mask(self):
+        lower, upper = np.reshape(
+            np.fromiter(
+                flatten(  # flatten the newly introduced (nan, nan) tuples
+                    replace(  # 'fixed' -> (nan, nan)
+                        flatten(self.hyperparameter_bounds),  # tree -> list
+                        'fixed',
+                        (np.nan, np.nan)
+                    )
+                ),
+                dtype=np.float
+            ),
+            (2, -1),
+            order='F'
+        )
+        inactive = np.isnan(lower) | np.isnan(upper) | (lower == upper)
+        return ~inactive
+
+    @property
+    def theta(self):
+        '''The logarithms of a flattened array of kernel hyperparameters,
+        excluing those declared as 'fixed' or those with equal lower and upper
+        bounds.'''
+        return np.log(self.flat_hyperparameters[self.active_theta_mask])
+
+    @theta.setter
+    def theta(self, value):
+        hypers = np.log(self.flat_hyperparameters)
+        hypers[self.active_theta_mask] = value
+        (self.p.theta,
+         self.q,
+         self.node_kernel.theta,
+         self.edge_kernel.theta
+         ) = fold_like(np.exp(hypers), self.hyperparameters)
+
     @property
     def bounds(self):
-        return np.log(np.fromiter(flatten(self.hyperparameter_bounds),
-                                  np.float)).reshape(-1, 2, order='C')
+        '''The logarithms of a reshaped X-by-2 array of kernel hyperparameter
+        bounds, excluing those declared as 'fixed' or those with equal lower
+        and upper bounds.'''
+        return np.log(
+            np.fromiter(
+                flatten(
+                    replace(
+                        flatten(self.hyperparameter_bounds),
+                        'fixed',
+                        (np.nan, np.nan)
+                    )
+                ),
+                np.float
+            ).reshape(-1, 2, order='C')[self.active_theta_mask, :]
+        )
 
     def clone_with_theta(self, theta):
         clone = copy.deepcopy(self)
         clone.theta = theta
         return clone
-
-    def get_params(self, deep=False):
-        return dict(
-            node_kernel=self.node_kernel,
-            edge_kernel=self.edge_kernel,
-            p=self.p,
-            q=self.q,
-            q_bounds=self.q_bounds,
-            backend=self.backend
-        )
